from __future__ import annotations

from dataclasses import dataclass

from dozersim.path import Path
from dozersim.modelling import Model, Settings
from dozersim import constraints


@dataclass(slots=True)
class SpringSettings(Settings):
    stiffness: float = None
    max_torque: float = None


def evaluate_spring_deflection(paths: tuple[Path, Path], spring: SpringSettings, model: Model):
    path = paths[0]

    force = path.variables.effort
    deflection = force / spring.stiffness
    path.variables.add('displacement', deflection, model, "Spring deflection")

    model.constraints['peak_torque'].evaluate(model=model, path=path)


def evaluate_spring_force(paths: tuple[Path, Path], spring: SpringSettings, model: Model):
    path = paths[0]

    flows = path.variables.flows
    displacement1 = flows[-1].data[0]
    displacement2 = flows[-2].data[0]
    deflection = displacement2 - displacement1
    force = deflection * spring.stiffness
    path.variables.add('effort', force, model, "Spring force")

    model.constraints['peak_torque'].evaluate(model=model, path=path)


class Spring(Model):

    def __init__(self, name: str = 'spring', settings=SpringSettings(), eval_fun=evaluate_spring_deflection):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='peak_torque', settings=settings, ub='max_torque',
                                                       statistic='absmax', target='effort'))



from __future__ import annotations
from sqlalchemy import Column, String, Integer, Float
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, Query
from sqlalchemy.orm import sessionmaker
import os
import pathlib
file_path = pathlib.Path(__file__)
db_path = file_path.parent.absolute().joinpath('../data/supplier_data.db')

if db_path:
    engine = create_engine('sqlite:///'+str(db_path))
    Session = sessionmaker(bind=engine)
    session = Session()
    Base = declarative_base()
else:
    raise Exception('The supplier database was not found')


class BatteryTable(Base):
    """docstring for ElectricMotorSettings.Objects"""
    __tablename__ = 'batteryspecification'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    capacity = Column(Float)
    peak_current = Column(Float)
    voltage = Column(Float)
    mass_spec = Column(Float)


class MotorTable(Base):
    """docstring for ElectricMotorSettings.Objects"""
    __tablename__ = 'motorspecification'

    id = Column(Integer, primary_key=True)
    built = Column(String)
    name = Column(String)
    windingResistance = Column(Float)
    speedConstant = Column(Float)
    maxSpeed = Column(Float)
    ratedPower = Column(Float)
    peakPower = Column(Float)
    ratedCurrent = Column(Float)
    peakCurrent = Column(Float)
    windingVoltage = Column(Float)
    rotorInertia = Column(Float)
    mass = Column(Float)
    motorLength = Column(Float)
    motorDiameter = Column(Float)

    
class BallBearingTable(Base):
    __tablename__ = 'bearingspecifications'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    built = Column(String)
    limitingSpeed = Column(Float)
    boreDiameter = Column(Float)
    outerDiameter = Column(Float)
    width = Column(Float)
    mass = Column(Float)
    staticLoadRating = Column(Float)
    dynamicLoadRating = Column(Float)
    minimumLoadFactor = Column(Float)
    loadRatioLimit = Column(Float)
    factorX = Column(Float)
    factorY0 = Column(Float)
    factorY2 = Column(Float)   


class GreaseTable(Base):
    __tablename__ = 'greasespecifications'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    oilType = Column(String)
    viscosityAt40Deg = Column(Float)
    viscosityAt100Deg = Column(Float)




from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
from dozersim.modelling import Model, Settings
from dozersim.path import Path


@dataclass(slots=True)
class GaussiansFilterSettings(Settings):
    Target: str = None
    Points: list = None
    PeakValue: float = None
    Width: float = None


def evaluate_gaussian_filter(paths: tuple[Path], profile: GaussiansFilterSettings, model: Model):
    path, = paths

    time = path.variables.time
    y_values = []
    width = profile.Width  # 100
    nSamples = len(time)#profile.NumberOfSamples
    nPoints = len(profile.Points)
    kernel = np.arange(nPoints)
    omega = time[-1] * 2 * np.pi
    phis = np.arange(nSamples)/nSamples*omega
    points = np.array(profile.Points)

    for phi in phis:
        q, dq, ddq = gaussians(phi, omega, kernel, width, points)
        y_values.append(q)

    path.variables.add(profile.Target, y_values, model, "Profile")


def gaussians(phi: float, omega: float, kernel_i: np.ndarray, width: float, weights: np.ndarray) -> tuple[float, float, float]:
    """
    % This function implements a discretized version of the adaptive non-linear filter of the adaptive oscillator
    % for non-sinusoidal profiles

     inputs:
       phi: the phase of the signal
       omega: the frequency of the signal (rad/s)
       kernel: ndarray: indeces of the gaussians (e.g. 1:20 for 20 gaussians)
       width: the width of the gaussians
       weights: ndarray the weight factors of the gaussians
     outputs:
       est_pos: the estimated position
       est_vel: the estimated velocity
       est_acc: the estimated acceleration

    REF e.g.
    [1] A. Gams, A. J. Ijspeert, S. Schaal, and J. Lenar?i?, “On-line
    learning and modulation of periodic movements with nonlinear dynamical
    systems,” Auton. Robots, vol. 27, no. 1, pp. 3–23, 2009.
    """
    c_kernel = kernel_i/len(kernel_i)*2*np.pi

    psi = np.exp(width * (np.cos(np.tile(phi, (len(c_kernel), 1)) - c_kernel) - 1))
    dpsi = -psi * width * np.sin(np.tile(phi, (len(c_kernel), 1)) - c_kernel) * omega
    ddpsi = -dpsi * width * np.sin(np.tile(phi, (len(c_kernel), 1))
                                   - c_kernel) * omega - psi * width * np.cos(np.tile(phi, (len(c_kernel), 1))
                                                                              - c_kernel) * omega ** 2
    # -psi*h.*sin(long_phase)*0

    est_pos = np.sum(psi * weights) / np.sum(psi)
    est_vel = np.sum(dpsi * weights) / np.sum(psi)
    est_acc = np.sum(ddpsi * weights) / np.sum(psi)
    return est_pos, est_vel, est_acc



class GaussianFilter(Model):

    def __init__(self, name: str = 'gaussian filter', settings=GaussiansFilterSettings(), eval_fun=evaluate_gaussian_filter):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)




from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
from dozersim.path import Path
from dozersim.modelling import Model, Settings
from dozersim.results import Load


@dataclass(slots=True)
class MultiplierSettings(Settings):
    number: int = None # Number of identical items
    target: str = None # e.g. 'effort' or 'flow'


def evaluate_multiplier(paths: tuple[Path], multiplier: MultiplierSettings, model: Model):
    """
        Multipliers simulates multiple identical items
        multipliers are independent of the type of variable they are applied to
        and can be used to multiply the results of a path by a constant factor.

    """
    path, = paths

    gains = {'flow': 1, 'effort': 1}
    try:
        gains[multiplier.target] = multiplier.number
    except KeyError:
        raise ValueError(f"Unknown multiplier target: {multiplier.target}. Available targets are: {list(gains.keys())}")
    
    path.transform(var_type=path.variables.domain, name=f'{multiplier.number} items in {multiplier.target}')

    for target in gains.keys():
        values = getattr(path.variables, target)
        path.variables.add(target=target,
            values=values*gains[target],
            parent=model,
            name=f'Multiplier {multiplier.target}')



class Multiplier(Model):
    def __init__(self, name: str = 'multiplier', settings=MultiplierSettings(), eval_fun=evaluate_multiplier):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)



from __future__ import annotations
from dataclasses import dataclass
import numpy as np
from dozersim.modelling import Settings
from dozersim.modelling import Model
from dozersim.path import Path
from dozersim.results import Load
from dozersim import constraints
from dozersim.components import tools


@dataclass(slots=True)
class BallScrewSettings(Settings):
    name: str = None
    mass: float = None
    pitch: float = None
    nominal_diameter: float = None
    dynamic_load_rating: float = None
    static_load_rating: float = None
    efficiency: float = None
    inertia: float = None

    @property
    def speed_limit(self):
        return 50000 / (self.nominal_diameter * 10e-3)

    @property
    def supplier_table(self):
        return None


def evaluate_ball_screw(paths: list[Path], ballscrew: BallScrewSettings, model: Model):
    """
        Evaluates a ball screw spindle transmission

    Parameters
    ----------
    paths: Path that will be used for all operation on the variables
    ballscrew: Settings of the linkage mechanism
    model: used to leave a parent reference in variables

    Returns
    -------

    """
    path, = paths
    path.check('translational')

    forces_out = path.variables.efforts
    force_out_total = path.variables.torque
    stroke_out = path.variables.displacement - np.min(path.variables.displacement)
    velocity_out = path.variables.flow
    acceleration_out = path.variables.acceleration

    path.transform(var_type='rotational', name="Ball screw input")

    ratio = 2 * np.pi / ballscrew.pitch

    time = path.variables.time
    speed_in = velocity_out * ratio
    acceleration_in = acceleration_out * ratio
    torque_inertia = acceleration_in * ballscrew.inertia
    n_rotations = np.trapz(np.abs(speed_in) * 30 / np.pi, time)
    t_rotation = time[-1]
    rotation_per_hour = n_rotations/t_rotation * 60 * 60
    speed_in_max = np.max(np.abs(speed_in))
    load_cubic_mean = np.sum(np.abs(force_out_total) ** 3 * stroke_out) ** (1 / 3) / np.sum(stroke_out) ** (1 / 3)
    load_static = np.max(np.abs(force_out_total))

    life_rating_10 = (ballscrew.dynamic_load_rating/load_cubic_mean) ** 3
    life_rating_10h = 10e6 / rotation_per_hour * life_rating_10

    path.variables.add('flow', speed_in, model, "Ball screw input")

    for force_out in forces_out:
        path.variables.add('effort',force_out.data / ratio, model, force_out.name)

    friction_torque = tools.calculate_friction_torque(effort=path.variables.effort, power=path.variables.power,
                                                      efficiency=ballscrew.efficiency)

    path.variables.add(target='effort', values=np.array(friction_torque), parent=model, name="Ball screw friction")
    path.variables.add(target='effort', values=torque_inertia, parent=model, name="Ball screw inertia")
    model.constraints['static_force'].evaluate(model=model, path=path)
    model.constraints['max_speed'].evaluate(model=model, path=path)
    path.add_result_object(Load(parent=model, path=path,
                                force_radial=np.full(np.asarray(force_out_total).shape, 0),
                                force_axial=force_out_total))


class BallScrew(Model):

    def __init__(self, name: str = 'ballscrew', settings=BallScrewSettings(), eval_fun=evaluate_ball_screw):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='static_force', settings=settings, ub='static_load_rating',
                                                       statistic='absmax', target='effort'))
        self.add_constraint(constraints.PathConstraint(name='max_speed', settings=settings, ub='speed_limit',
                                                       statistic='absmax', target='effort'))



from __future__ import annotations
from dataclasses import dataclass
import numpy as np

from dozersim import constraints
from dozersim.suppliers import MotorTable
from dozersim.path import Path
from dozersim.modelling import Settings
from dozersim.modelling import Model


@dataclass(slots=True)
class ElectricMotorSettings(Settings):
    name: str = None
    built: str = None
    windingResistance: float = None
    speedConstant: float = None
    maxSpeed: float = None
    peakPower: float = None
    ratedPower: float = None
    ratedCurrent: float = None
    peakCurrent: float = None
    windingVoltage: float = None
    rotorInertia: float = None
    mass: float = None
    motorLength: float = None
    motorDiameter: float = None

    @property
    def torqueConstant(self):
        return 1 / (self.speedConstant * np.pi / 30)

    @property
    def supplier_table(self):
        return MotorTable


def evaluate_electric_motor(paths: tuple[Path], motor: ElectricMotorSettings, model: Model):
    path = paths[0]

    path.check('rotational')

    # Get speed and
    speed = path.variables.velocity
    acceleration = path.variables.acceleration

    # Calculate and add rotor inertia moment
    if motor.rotorInertia:
        moment_inertia = motor.rotorInertia * acceleration
        path.variables.add('torque', moment_inertia, model, "Motor Inertia")

    # Calculate motor current and voltage
    torque = path.variables.torque
    motor_current = torque / motor.torqueConstant
    emf_voltage = speed * motor.torqueConstant
    resistance_voltage = motor.windingResistance * torque / motor.torqueConstant
    motor_voltage = emf_voltage + resistance_voltage
    motor_power = motor_voltage * motor_current

    path.transform(var_type='electrical', name="Motor input")
    path.variables.add('current', motor_current, model, "Motor Current")
    path.variables.add('voltage', emf_voltage, model, "EMF Voltage")
    path.variables.add('voltage', resistance_voltage, model, "Resistance Voltage")

    model.constraints['rated_current'].evaluate(model=model, path=path)
    model.constraints['peak_current'].evaluate(model=model, path=path)
    model.constraints['voltage'].evaluate(model=model, path=path)
    model.constraints['rated_power'].evaluate(model=model, path=path)
    model.constraints['peak_power'].evaluate(model=model, path=path)


class ElectricMotor(Model):
    def __init__(self, name: str = 'electricmotor', settings=ElectricMotorSettings(), eval_fun=evaluate_electric_motor):
        super().__init__(name=name, settings=settings, eval_fun=eval_fun)
        self.add_constraint(constraints.PathConstraint(name='rated_current', settings=settings, ub='ratedCurrent',
                                                       statistic='absmean', target='flow'))
        self.add_constraint(constraints.PathConstraint(name='peak_current', settings=settings, ub='peakCurrent',
                                                       statistic='absmax', target='flow'))
        self.add_constraint(constraints.PathConstraint(name='voltage', settings=settings, ub='windingVoltage',
                                                       statistic='absmax', target='effort'))
        self.add_constraint(constraints.PathConstraint(name='rated_power', settings=settings, ub='ratedPower',
                                                       statistic='absmean', target='power'))
        self.add_constraint(constraints.PathConstraint(name='peak_power', settings=settings, ub='peakPower',
                                                       statistic='absmax', target='power'))



"""
This module defines the `Model` class, an abstract base class representing
a simulation model with configurable settings, evaluation functions, and constraints.

Classes
-------
Model
    Abstract base class for creating simulation models.
"""
from __future__ import annotations
from abc import ABC
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dozersim.modelling import Settings
    from dozersim.path import Path
    from dozersim.constraints import Constraint


class Model(ABC):
    """Abstract base class representing a simulation model.

    Attributes
    ----------
    name : str
        Name of the model.
    _settings : Settings
        Simulation settings object.
    _eval_fun : callable
        Evaluation function to run simulations.
    post_eval : list[callable]
        List of post-evaluation functions.
    _constraints : dict[str, Constraint]
        Dictionary storing constraints by name.
    """

    def __init__(self, eval_fun: callable, name: str, settings: Settings) -> None:
        """Initializes the Model with evaluation function, name, and settings.

        Parameters
        ----------
        eval_fun : callable
            Function used to evaluate paths.
        name : str
            Name of the model.
        settings : Settings
            Simulation settings object.
        """
        self.name = name
        self._settings = settings
        self._eval_fun = eval_fun
        self.post_eval: list[callable] = []
        self._constraints: dict[str, Constraint] = {}

    def __repr__(self) -> str:
        """Returns a string representation of the Model.

        Returns
        -------
        str
            A representation string showing the model's name and settings.
        """
        return f'settings.Lead("{self.name}", function, settings)'

    def __str__(self) -> str:
        """Returns the model's name as a string.

        Returns
        -------
        str
            The model's name.
        """
        return self.name

    @property
    def settings(self) -> Settings:
        """Gets the current simulation settings.

        Returns
        -------
        Settings
            The current simulation settings object.
        """
        return self._settings

    @settings.setter
    def settings(self, settings: Settings):
        """Sets new simulation settings.

        Parameters
        ----------
        settings : Settings
            New simulation settings object.

        """
        self._settings = settings

    def evaluate(self, paths: list[Path]):
        """Evaluates paths using the evaluation function and settings.

        Parameters
        ----------
        paths : list[Path]
            List of paths to evaluate.
        """
        self._eval_fun(paths, self._settings, self)
        for fun in self.post_eval:
            fun(paths)

    def add_constraint(self, constraint: Constraint):
        """Adds a constraint to the internal constraint dictionary.

        Parameters
        ----------
        constraint : constraints.Constraint
            The constraint to add. Its name is used as the dictionary key.
        """
        cons_name = constraint.name
        self._constraints[cons_name] = constraint

    @property
    def constraints(self):
        """Retrieves all constraints in the internal dictionary.

        Returns
        -------
        dict[str, constraints.Constraint]
            Dictionary mapping constraint names to `Constraint` objects.
        """
        return self._constraints



from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class Material:
    name: str = None
    density: float = None
    youngsModulus: float = None
    yieldStrength: float = None
    poissonsRatio: float = None
    contactStrength: float = None


from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

import numpy as np

from dozersim.utils import target_dict, statistic_dict

if TYPE_CHECKING:
    from dozersim.modelling import Model
    from dozersim.path import Path

from dozersim.results import ObjectiveValue, ConstraintValue


class Objective(ABC):

    def __init__(self, statistic: str, goal: str,
                 scale: float = 1) -> None:
        self.statistic = statistic_dict[statistic]
        self.goal = goal_dict[goal]
        self.scale = scale
        self._optimize = True

    @abstractmethod
    def engage(self):
        pass

    @abstractmethod
    def disengage(self):
        pass

    @abstractmethod
    def evaluate(self, paths: tuple[Path]):
        pass

    @property
    def name(self):
        pass

    @property
    def optimize(self):
        pass

    @property
    @abstractmethod
    def trace(self):
        pass


class SettingsObjective(Objective):
    """
    The SettingsObjective class is and implementation of Object that collects settings with a certain keyword.

    """

    def __init__(self, models: Model | list[Model], statistic: str, goal: str,
                 target: str, scale: float = 1) -> None:
        super().__init__(statistic, goal, scale)
        if type(models) is not list: models = [models]
        self._models = models
        self.target = target
        self._optimize = False

    @property
    def name(self):
        return f'{self.statistic.name()} {self.target}'

    @property
    def trace(self):
        return self._models.name

    @property
    def optimize(self):
        return self._optimize

    @optimize.setter
    def optimize(self, do_optimize: bool):
        if do_optimize:
            raise Warning("Dragons ahead! Setting a SettingsObjective to be optimized, can cause unstable behavior "
                          "since settings objective can be highly discontinuous.")
        self._optimize = do_optimize

    def engage(self):
        self._models[-1].post_eval.append(self.evaluate)

    def disengage(self):
        self._models[-1].post_eval.remove(self.evaluate)

    def evaluate(self, paths):

        setting_values = []

        if isinstance(self._models, list):
            models = self._models
        elif isinstance(self._models, Model):
            models = [self._models]
        else:
            raise Exception('Component object in objective function is not an instance of model or list of models')

        for model in models:
            try:
                value = model.settings.get_setting(self.target)
            except AttributeError:
                pass
            else:
                setting_values.append(value)

        if len(setting_values) == 0:
            raise Exception(
                f'Setting {self.target} not found in object-scope. Please check items of remove objective')

        cost_value = self.statistic.evaluate(np.array(setting_values))
        paths[0].add_result_object(
            ObjectiveValue(parent=self, unit='', value=cost_value, path=paths[0]))


class ConstraintObjective(Objective):
    """
    The Constraint Objective class is an implementation of Objective
    """

    def __init__(self, model: Model, goal: str,
                 target: str, scale: float = 1) -> None:
        super().__init__('None', goal, scale)
        self._model = model
        self._target = target
        self._optimize = False

    @property
    def name(self):
        return f'{self._model.name}.{self._target}'

    @property
    def trace(self):
        return self._model.name

    @property
    def optimize(self):
        return self._optimize

    @optimize.setter
    def optimize(self, do_optimize: bool):
        self._optimize = do_optimize

    def engage(self):
        self._model.post_eval.append(self.evaluate)

    def disengage(self):
        self._model.post_eval.remove(self.evaluate)

    def evaluate(self, paths: tuple[Path] = ()):
        for path in paths:
            const_objs: list[ConstraintValue] = path.get_result_object(ConstraintValue)
            for const_obj in const_objs:
                if const_obj.model == self._model and const_obj.parent.name == self._target:
                    cost_value = const_obj.value
                    path.travelers.remove(const_obj)
                    path.add_result_object(
                        ObjectiveValue(parent=self, unit=const_obj.unit, value=cost_value, path=path))


class PathObjective(Objective):

    def __init__(self, model: Model, statistic: str, target: str, goal: str,
                 scale: float = 1, path: Path = None) -> None:
        super().__init__(statistic, goal, scale)
        self.target = target_dict[target]
        self._model = model
        self._path = path
        self._optimize = True

    @property
    def name(self):
        return f'{self._model.name}.{self.statistic.name()} {self.target.name()}'

    @property
    def trace(self):
        return self._model.name

    @property
    def optimize(self):
        return self._optimize

    @optimize.setter
    def optimize(self, do_optimize: bool):
        self._optimize = do_optimize

    def engage(self):
        self._model.post_eval.append(self.evaluate)

    def disengage(self):
        self._model.post_eval.remove(self.evaluate)

    def evaluate(self, paths: tuple[Path]):
        if self._path and (self._path in paths):
            path = self._path
        else:
            path, = paths

        cost_values = self.target.evaluate(path)
        cost_value = self.statistic.evaluate(cost_values)

        path.add_result_object(
            ObjectiveValue(parent=self, unit=self.target.units(path), value=cost_value, path=path))


goal_dict = {'maximize': -1,
             'minimize':  1}
